import itertools
from dataclasses import dataclass
from pathlib import Path
from typing import NoReturn, Optional, Union

from statute_api.config import (
    data_tree_walker,
    format_date,
    load_yaml_from_path,
)
from statute_api.unit_maker import format_units


@dataclass
class CodificationItem:

    path_to_code_yaml: Path

    def __post_init__(self):
        self.data: dict = self.complete_data
        self.sync = self.statutes_in_sync
        self.units: list[dict] = self.get_units
        self.category = "COD"  # for codification
        self.version = str(self.data["version"])
        self.email = self.data["email"]
        self.title = self.data["title"]
        self.description = self.data["description"]
        self.base = self.data["base"]
        self.histories: list[dict] = self.get_histories

    """Post_initialize checking based on data contained in location."""

    @property
    def complete_data(self) -> Union[NoReturn, dict]:
        data = load_yaml_from_path(self.path_to_code_yaml)
        ks = ["title", "description", "email", "version", "base", "requisites"]
        if missing_fields := [k for k in ks if k not in data]:
            raise Exception(f"{missing_fields}")
        return data

    @property
    def get_units(self) -> Optional[list]:
        if not (units_found := self.data.get("units", None)):
            raise Exception(f"No units from {self.data}")
        format_units(units_found)
        return units_found

    @property
    def statutes_from_histories(self) -> Union[set[str], NoReturn]:
        """Get unique statutes"""
        if not (blocks := set(data_tree_walker(self.data, "statute"))):
            raise Exception(f"No statute blocks found; see {blocks=}")
        return blocks

    @property
    def requisites_sans_base(self) -> set[str]:
        return set(self.data["requisites"])

    @property
    def statutes_in_requisites(self) -> set[str]:
        return set(self.data["requisites"] + [self.data["base"]])

    @property
    def no_reqs(self) -> set[str]:
        return self.statutes_from_histories - self.statutes_in_requisites

    @property
    def no_hist(self) -> set[str]:
        return self.statutes_in_requisites - self.statutes_from_histories

    @property
    def statutes_in_sync(self) -> Union[bool, NoReturn]:
        if self.no_reqs:
            raise Exception(f"In body not in requisites: {str(self.no_reqs)=}")
        if self.no_hist:
            raise Exception(f"In requisites not in body: {str(self.no_hist)=}")
        return True

    """Assign attributes post-init"""

    @property
    def publication_date(self):
        return format_date(x) if (x := self.data.get("date", None)) else None

    @property
    def get_history_lists(self) -> Union[list[list[dict]], NoReturn]:
        """Each unit will have its own history; each history is a list of blocks"""
        if not (l := list(data_tree_walker(self.data, "history"))):
            raise Exception(f"No list of history blocks found; see {l=}")
        return l

    @property
    def get_histories(self) -> Union[list[dict], NoReturn]:
        """Get unique histories"""
        result = itertools.chain(*self.get_history_lists)  # flattens list
        return list(result)
